#!/usr/bin/perl -w -*- perl -*-
# Program written by: Jens Heunemann
# (c) Jens Heunemann
#
# DATE 20.04.2001
# Version 1.02
#
# This program is protected by Copyright-law.
# Unauthorised use prohibited
# If you desire to use it, contact
# Jens Heunemann
# jens@consol.de
#
# changes 1.02 (Thomas Linden) [20.04.2001]:
#  - added array support in add_code which helps
#    avoiding huge nested template constructs (loops)
#  - added pod for this
#  - pod updated for readability
#  - removed debug code

package Template;

use 5.004;

$Template::VERSION='1.02'; # ==> ALSO Update POD!!!

use Exporter ();
use FileHandle ();
use strict;
use Carp;
use constant DEBUG => 0;
use File::Copy qw(move);

@Template::ISA=qw(Exporter);
@Template::EXPORT=qw();
@Template::EXPORT_OK=qw();
@Template::EXPORT_TAGS=();

*code=\&add_code;
*print=\&template_print;
*values=\&add_values;

1;


# new Method for Template
# Entweder Parameter als
# -name=>wert,-name2=>wert
# bergeben,
# oder in der Reihenfolge
# von @order

# template: String mit dem Template, entweder mit $t->template("...") erzeugen
# oder mit $t->read einlesen
# real_name: Name des Ziel-files
# template_name: Name des Template-files
# code: ref auf eine Unterroutine, die in $_ den Namen der Funktion kriegt,
#       und als Parameter die Parameter an die Funktion

sub new {
  my $class=shift;
  my $type=ref($class)||$class;
  my $self={"code"=>[],
	    "error_templates"=>[],
	    "type"=>"advanced",
	   };
  bless $self,$type;
  return $self unless @_>0;

  #Named Parameter (mit - davor) oder nur ein Parameter?
  if (substr($_[0],0,1) eq "-") {
    my ($key,$wert);
    while (scalar @_>1) {
      $key=shift;
      $key =~ s/^-// or croak "Named Parameter for Template must begin with a minus";
      if ($key eq "code") {
	my $code=shift;
	croak "Parameter -code must be a hashreference"
	  unless ref($code)=~/HASH/;
	$self->add_code(%{$code});
      }
      elsif ($key eq "name") {
	my $name=shift;
	croak "Template $name does not exist or is not readable"
	  unless -e $name && -r $name;
	$self->{$key}=$name;
      }
      elsif ($key eq "values") {
	my $values_ref=shift;
	croak "Parameter -values must be a hashreference"
	  unless ref($values_ref)=~/HASH/;
	$self->add_values($values_ref);
      }
      else {
	$self->{$key}=shift;
      }
    }
    croak "Odd number of Named Parameters for Template" if scalar @_ == 1;
  }
  else {
    my $name=shift;
    croak "Template $name does not exist or is not readable"
      unless -e $name && -r $name;
    $self->{"name"}=$name;
  }
  return $self;
}

sub template_name {
  my $self=shift;
  if (@_==0) {
    return $ {$self}{'name'};
  }
  else {
    my $name=shift;
    croak "Template $name does not exist or is not readable"
      unless -e $name && -r $name;
    return $ {$self}{'name'}=$name;
  }
}

sub real_name {
  my $self=shift;
  if (@_==0) {
    return $ {$self}{'real_name'};
  }
  else {
    my $name=shift;
    croak "Template-target or Backupfile is not writable or cannot be created"
      unless ((-e $name and -w $name or
	       open(T,">$name") and close(T)) and
	      (-e "$name.bak" and -w "$name.bak"
	       or open(T,">$name.bak") and close(T)));
    return $ {$self}{'real_name'}=$name;
  }
}

sub template {
  # template setzen
  my $self=shift;
  if (@_==0) {
    return \ ($self->{'template'});
  }
  else {
    return $ {$self}{'template'}=shift;
  }
}

sub add_code {
  # code wird intern als liste
  # [[regexp1,coderef1],[regexp2,coderef2],...]
  # gespeichert
  my $self=shift;
  while (@_>1) {
    my $regexp=shift;
    my $code=shift;
    if (ref($code)=~/CODE/) {
      unshift @{$self->{"code"}},[$regexp,$code];
    }
    elsif (ref($code)=~/SCALAR/) {
      my $wert=$$code;
      unshift @{$self->{"code"}},[$regexp,sub { $wert }];
    }
    elsif (ref($code)=~/HASH/) {
      my %hash=%$code;
      unshift @{$self->{"code"}},[$regexp,
				  sub {
				    my $par=$_[-1];
				    $par=~s/\[(.*?)\]/$hash{$1}/g;
				    $par;
				  }];
    }
    elsif (ref($code) =~ /ARRAY/) {
      my @array = @{$code};
      unshift @{$self->{"code"}},[$regexp,
				  sub {
				    my $par=$_[-1];
				    my $out;
				    foreach my $elem (@array) {
				      next if ref($elem) !~ /HASH/;
				      my $part = $par;
				      $part =~ s/\[(.*?)\]/$elem->{$1}/g;
				      $out .= $part;
				    }
				    $out;
				  }];
    }
    elsif (ref($code)) {
      croak "Error in set_code, wrong parameters:".ref($code);
    }
    else {
      unshift @{$self->{"code"}},[$regexp,sub { $code }];
    }
  }
  if (@_==1) {
    croak "Wrong number of parameters to set_code";
  }
}

sub add_values {

  my $self=shift;

  if (ref($_[0]) =~ /HASH/) {
    my $values_ref=shift;
    for (keys %$values_ref) {
      $self->add_code("^\Q$_\E\$"=>sub { return $values_ref->{$_[1]} },
		     );
    }
  }
  else {
    eval {
      my %hash=@_;
      for (keys %hash) {
	$self->add_code("^\Q$_\E\$"=>$hash{$_},
		       );
      }
    };
    if ($@) {
      die "Fehler bei der Parameterbergabe beim Aufruf von add_values";
    }
  }

}

sub add_error {
  # Fehlertemplates setzen
  # zu jedem errorcode wird ein templatename gespeichert.
  my $self=shift;
  while (@_>1) {
    my $regexp=shift;
    my $name=shift;
    croak "Error-Template does not exist or is not readable"
      unless -e $name && -r $name;
    unshift @{$self->{"error_templates"}},[$regexp,$name];
  }
  if (@_==1) {
    croak "Wrong number of parameters to set_code";
  }
}

sub error {
  # Mit parameter: Fehlercode setzen
  # Ohne Parameter: Fehlermeldung (aus Template) zurckgeben
  my $self=shift;
  if (@_==0) {
    my $t=new Template;
    my $ref;
  ERROR:
    for $ref (@{$self->{"error_templates"}}) {
      if ($self->{"error"} =~ /$ref->[0]/) {
	$t->template_name($ref->[1]);
	last ERROR;
      }
    }
    $t->print;
    return 1;
  }
  else {
    return $self->{"error"}=shift;
  }
}

# noch unschn, im Moment wird noch das Template komplett eingelesen,
#
sub read {
  my $self=shift;
  if (open TEMPLATE,"<$ {$self}{'name'}") {
    $ {$self}{'template'}="";
    while (<TEMPLATE>) {
      $ {$self}{'template'}.=$_;
    }
    close TEMPLATE;
  }
  else {
    croak "Could not open TEMPLATE: $!";
  }
  return $self;
}

sub write {
  my $self=shift;
  $self->read() unless defined $ {$self}{'template'};
  my $fh=new FileHandle;
  if ($fh->open(">$ {$self}{'real_name'}.bak")) {
    $self->print($fh) or return;
    $fh->close;
    move("$ {$self}{'real_name'}.bak",$ {$self}{'real_name'}) or
      croak "Error while moving $ {$self}{'real_name'}.bak to $ {$self}{'real_name'}: $!";
  }
  else {
    croak "Error while opening $ {$self}{'real_name'}.bak for writing: $!";
  }
  return $self;
}

sub template_print {
  my $self=shift;
  my $fh;
  if (@_>0) {
    $fh=shift;
  }
  else {
    $fh=\*STDOUT;
  }
  print $fh $self->process;
  return $self;
}

sub process {
  my $self=shift;
  my $ref;
  my $regexp;
  my $pos=0;
  my $ergebnis="";
  $self->read() unless defined $ {$self}{'template'};

  if ($self->{type} eq "simple") {
    while ($self->{template} =~
	   m/\G(.*?) #Klammer1
	     \[([^]]*)\]/igxs
	  ) {
      $pos=pos($self->{"template"});
      $ergebnis.=$1;
      my $name=$2;
      my $found=0;
    REGEXP:
      for $ref (@{$self->{"code"}}) {
	$name =~ /$ref->[0]/ and do {
	  $found=1;
	  my $value;
	  $value=$ref->[1]->($self,$name);

	  # Keine Rckgabe von der Unterroutine? dann auch nichts ans Ergebnis dranhngen
	  last REGEXP unless defined $value;

	  if (ref($value) =~ /SCALAR/) {
	    $ergebnis.=$$value;
	  }
	  else {
	    $ergebnis.=$value;
	  }
	  last REGEXP
	}
      }
      $ergebnis.="[$name]" unless ($found);
    }
  }
  else {
  MATCH:
    while ($self->{"template"} =~
	   m/\G(.*?)                          # Klammer1
	     <!--perl                         # Anfang des bots
	     (?:
	     :                                # Klammer- -  perl: dann bot ohne Inhalt
	     [\s\n]*?                         # beliebige leerzeichen
	     ([a-zA-Z0-9_]+)                  # Klammer2 -  Aufzurufende Funktion
	     (?:[\s\n]+(?=[^>\s\n])([^>]*?))? # Klammer3 - Evtl noch Parameter dazu
	     [\s\n]*?-->                      # Ende des bots
	     |
	   start:                             # Klammer- - perlstart: dann bot mit Inhalt
	     [\s\n]*?
	     ([a-zA-Z0-9_]+)                  # Klammer4 - Aufzurufende Funktion
	     (?:[\s\n]+(?=[^>\s\n])([^>]*?))? # Klammer5 - Evtl noch Parameter dazu
	     [\s\n]*?                         # Hier keine Parameter
	     -->
	     (.*?)                            # Klammer6 - Inhalt
	     <!--perlend(?::\s*\4\s*)?-->
	    )
	     /ixgs) {
      $pos=pos($self->{"template"});
      $ergebnis.=$1;
      my $name;
      my @par;
      if ($2) {
	$name=$2;
	@par=split " ",($3||"");
	print join "--",@par if DEBUG;
      }
      else {
	$name=$4;
	@par=split " ",($5||"");
	push @par,$6;
      }
    REGEXP:
      for $ref (@{$self->{"code"}}) {
	$name =~ /$ref->[0]/ and do {
	  my $value;
	  $value=$ref->[1]->($self,$name,@par);

	  # Keine Rckgabe von der Unterroutine? dann auch nichts ans Ergebnis dranhngen
	  last REGEXP unless defined $value;

	  if (ref($value) =~ /SCALAR/) {
	    $ergebnis.=$$value;
	  }
	  else {
	    $ergebnis.=$value;
	  }
	  last REGEXP
	}
      }
    }
  }
  $ergebnis.=substr($self->{'template'},$pos);
  return $ergebnis;
}






=head1 NAME

Template - Module to use powerful template, e.g. for html

=head1 SYNOPSIS

 use Template;
 my $t=new Template;               # create new Template-object
 $t->template_name("abc.tpl");     # template is loadad from file abc.tpl

 $t->code(                         # set code-refs for embedded functions
         "^greeting"  => \&greeting,
         "^time"      => sub{ return &createdate(time) },
         "^count"     => $count
 );
 $t->print;                        # print template

or alternatively:

 my $t2=Template->new(
                      -type     => "simple",
		      -template => "abc [xyz] [f]def [a]x blub",
   		      -values   => { "xyz" => 1, a=>10,},
	             );
 $t2->print

=head1 ABSTRACT

With this module you can create one or more (Text-)Templates. You can
embed html-comments with a special syntax into the Template, which are
replaced by values provided by subroutines you give as
code-references.  This way you can separate the text from your
perl-programm, making both easier to create and maintain.  This is
especially useful if you have Frontpage, because the html-comments
just happen to coincide with the comments you can embed into your html
with Frontpage.

But of course this module is not limited to html code, you can use it for
any other type of text too!


=head1 ADVANCED USAGE

One tricky thing about these templates is how to use a template in a template
or in other words, how to re-use a particular part of a template
inside a foreach() loop.

I.e. let's take an array of hashes, which should be displayed in a table:

 @array=(
          { "name" => "tobi" ,     "tel" => "123" } ,
          { "name" => "katarina" , "tel" => "456" } ,
          { "name" => "cornelia" , "tel" => "789" } ,
       );

Suppose you have the following template:

 <table>
  <tr><th>NAME</th><th>TELEFONE</th></tr>
  <!--perlstart: table-->
  <tr>
    <td><!--perl: name--></td>
    <td><!--perl: tel--></td>
  </tr>
  <!--perlend: table-->
  </table>

now, you can process this template  using closures:

 my $t=Template->new("filename_of_template.tpl");
 $t->code( "table"=>
  sub {
    $par=$_[2];               # get text of inner template (see SUB ARGS below!)
    my $t=Template->new;      # inner template, without filename
    $t->template($par);       # fill inner Template with text
    my %data;                 # this hash MUST be defined before the code-section
    my $erg="";               # this is the string which is returned in the end
    $t->code(                 # arbitrary data
      "."=> sub {             # is processed by this subroutine
        my $name=$_[1];       # name of the section, in this case either
                              # "name" or "tel"
        return $hash{$name};  # just return the contents of the hash
      },
    );
    # now we are doing our loop through our @array:
    foreach my $ref (@array) {
      %data=%$ref;            # do not use B<my> in front of %data here!
      $erg.=$t->process;      # each subsequent $t->process uses the momentary
                              # content of %data, so each gives back the next tablerow
    }
    return $erg;
  },
 );
 $t->print;

If you don't want to do special operations on the array before insertion
you can also supply it directly to the module, but the template file must
be modified a little bit:

 <table>
  <tr><th>NAME</th><th>TELEFONE</th></tr>
  <!--perlstart: table-->
  <tr>
   <td>[name]</td>
   <td>[tel]</td>
  </tr>
  <!--perlend: table-->
 </table>

Now you can use the following code to fill it at once (given the array
defined above @array):

 my $t=Template->new("filename_of_template.tpl");
 $t->code(
   "table" => \@array
         );
 $t->print;


=head1 DESCRIPTION

Methods:


=item new ($ )

create new Template-object and provide optional Filename

Parameters:
either just provide a filename or use named parameters:

 -name:     Filename
 -template: Template-text
 -code:     reference to a hash which contains the parameters
            of a call to the code-method
 -values:   a reference to a hash which contains plain values
            useful for -type=>simple templates
 -type:     either simple or advanced
            advanced templates use <!--perl: ...--> and <!--perlstart: ...-->
            simple templates just use [abc] and [def]
            see synopsis for an example

=item template_name ($ )

  provide Filename for Template

=item real_name ($ )

  Name of the File where the resulting text (after template-replaces)
  is saved

=item template ($ )

  Text for the template, if you don\'t load it from a file

=item add_code

=item code (@)

@ is split into RegExp/Code-pairs, which are added to the Template. e.g.:

  my $t=new Template;
  $t->code("regexp1"=>\&sub1,
	   "regexp2"=>\&sub2);

Each closure (anonymous sub-routine, i.e. "\&sub1") gets any array of
parameters, see SUB ARGS for more details.

=item add_error

   Add Error-Template-name

=item error(errorstring)

   Set internal errorstring

=item error

   Print Error-Template according to add_error and errorstring

=item read

   Read Template from file provided by B<template_name>

=item write

   If no Template-text is provided, B<read> the Template.
   Write processed Template to file provided by B<real_name>.

=item print

=item print(Filehandle)

   Print processed Template to STDOUT or Filehandle

=back


=head1 SUB ARGS

Subroutines passed to the module using the B<code()> method get
3 arguments if they are called:

  my $t=new Template;
  $t->code("regexp1" => sub {
                              my($this, $name, $text) = @_;
                              ...
                            });

The first one B<$this> is the current Template object itself, in most
(if not all) cases you will not need to use this.

The second one, B<$name> is the name of the current key, on which
the regexp has matched. This is quite interesting if you had an expression
like this one:

 $t->code(
           "^master_\d{6}" => sub { ... },
         );

As you can see, the above sub would be executed on "master_123123" or
on "master_000000", but you cannot know, but inside the sub, you will
know, since the variable B<$name> tells you.

The last, and most important one, is B<$text>, which contains the junk
of text for which the expression had matched. I.e.:

 <table>
  <tr><th>NAME</th><th>TELEFONE</th></tr>
  <!--perlstart: table-->
  <tr>
   <td>[name]</td>
   <td>[tel]</td>
  </tr>
  <!--perlend: table-->
 </table>

If you use an expression of "table" then the variable B<$text> (the 3rd ARG)
to the sub will be:

  <tr>
   <td>[name]</td>
   <td>[tel]</td>
  </tr>

Inside your sub you can do with it what ever you want.

=head1 TEMPLATE

Html-comments of one of the following types are replaced by the values provided
by the code you gave with B<code>:

nearly uptodate:

 <!--perl:\s*FUNCTION PARAMETER1 PARAMETER2\s*-->

or

 <!--perlstart:\s*FUNCTION PARAMETER1 ...\s*-->
   LAST PARAMETER
 <!--perlend: FUNCTION-->

or
 <!--perlstart:\s*FUNCTION PARAMETER1 ...\s*-->
   some text and some [keys] for [replacing].
 <!--perlend: FUNCTION-->

FUNCTION is matched against the provided RegExps, matching
LastInFirstToMatch, for the first matching RegExp the corresponding
code-ref is called with the Template-object as first parameter (so the
code can provide error-messages), the name of FUNCTION and
PARAMETER1..x as the other Parameters.

The sub should return undef if an error occurs, and non-undef when successful.
If an error occurs, the according error-Template is printed.

In the SECOND form (with perlstart and perlend) LAST PARAMETER is the
whole string between.

=head1 AUTHOR INFORMATION

Copyright 1998, Jens Heunemann. All rights reserved.

=head1 BUGS

This is very alpha, the main thing works, but the interface will probably reworked
a number of times.

=cut
